from .base import BaseModel


class DateRange(BaseModel):
    _date_fields = ['start_date', 'finish_date']

    def _fill_date_fields(self, data):
        for field in self._date_fields:
            if field not in data:
                data[field] = None
        return super(DateRange, self)._fill_date_fields(data)


class Price(BaseModel):
    _as_is_fields = ['currency']
    _price_fields = ['amount', 'deposit']

    @property
    def _model_collection_fields(self):
        from .price_promotion import PricePromotion
        return [('promotions', PricePromotion)]


class PriceBand(BaseModel):
    _as_is_fields = [
        'code', 'max_age', 'max_travellers', 'min_age',
        'min_travellers', 'name',
    ]

    @property
    def _model_collection_fields(self):
        return [('prices', Price)]

    def available_currencies(self):
        return self.prices.keys()


class SeasonalPriceBand(PriceBand):

    @property
    def _model_collection_fields(self):
        return super(SeasonalPriceBand, self)._model_collection_fields + [
            ('season_dates', DateRange),
            ('blackout_dates', DateRange),
        ]


class PP2aPrice(BaseModel):
    _as_is_fields = ['currency']
    _price_fields = ['amount']



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from functools import partial
from importlib import import_module

def get_resource_class_from_class_name(name):
    resource_module = import_module('gapipy.resources')
    return getattr(resource_module, name)


def get_resource_class_from_resource_name(name):
    mapping = {resource._resource_name: resource
               for resource in get_available_resource_classes()}
    return mapping[name]


def get_available_resource_classes():
    from .resources import available_resources
    resource_module = import_module('gapipy.resources')
    return [getattr(resource_module, r) for r in available_resources]

def dict_to_model(class_name=None):
    """
    Wrapper to be used within resource definitions for objects you'd like to be
    dot-accessible without explicitly defining model fields for any nested data.

    You should use this with `_model_fields`, or `_model_collection_fields`,
    like so:

    _model_collection_fields = [
        ('phone_numbers', dict_to_model('Phone Numbers')),
    ]

    The optional`class_name` simply improves readability of the object
    within a debug environment.
    """
    from .models.base import DictToModel
    return partial(DictToModel, class_name=class_name)

def is_free(amount):
    """
    Explit zero amounts are interpreted as Free!
    """
    return (
        amount == 0 or
        amount == 0.00 or
        amount == '0' or
        amount == '0.00'
    )


def humanize_amount(amount, force_decimal=False):
    """
    Takes an `amount` (float) and removes any unnecessary decimals,
    or adds them if there is any partial amount.

    TODO: Internationalization support
    """
    if is_free(amount):
        return 'Free'

    amount = float(amount)
    if amount % 1 or force_decimal:
        return '%.2f' % amount
    return '%.0f' % amount


def humanize_price(amount_min, amount_max, currency):
    """
    Format a single price or price range for display.
    """
    # No price, nothing to display
    if amount_min is None:
        return ''

    # Price Range
    if amount_max:
        human_min = humanize_amount(amount_min)
        # If the miniumum has a decimal, then we want the max
        # to also have decimal places.
        force_decimal = '.' in human_min
        return '{}-{}{}'.format(
            human_min,
            humanize_amount(amount_max, force_decimal=force_decimal),
            currency,
        )
    # Single Price
    return '{}{}'.format(
        humanize_amount(amount_min),
        currency if not is_free(amount_min) else '',
    )


def humanize_time(hours):
    """
    Make a friendly duration label for display. e.g. Takes a value
    like 2.75 and returns 2h45m.
    """
    if not hours:
        return ''
    # TODO: Why are durations strings in our API?
    hours = float(hours)
    minutes = int(hours % 1 * 60)
    whole_hrs = int(hours)
    if minutes and whole_hrs:
        return '{}h{}m'.format(whole_hrs, minutes)
    elif minutes:
        return '{}m'.format(minutes)
    elif whole_hrs:
        return '{}h'.format(whole_hrs)


def location_label(start, end):
    """
    `start` and `end` are to locations that have a name property
    """
    if not start:
        return ''

    if end and end.name != start.name:
        return '{} – {}'.format(
            start.name,
            end.name,
        )
    return start.name


def duration_label(min_hr, max_hr):
    """
    Helper to output a friendly duration single value or range.
    """
    if not min_hr:
        return ''
    if max_hr:
        return '{}-{}'.format(humanize_time(min_hr), humanize_time(max_hr))
    return humanize_time(min_hr)


class LocationLabelMixin(object):
    """
    Mixin for resources with `start_location` and `end_location` fields.
    Formats a friendly label like Toronto - Monteal, or just a single
    location if both start and end are the same.
    """
    @property
    def location_label(self):
        return location_label(self.start_location, self.end_location)


class DurationLabelMixin(object):
    """
    Mixin for resources with a duration dict to format a human friendly
    single or ranged duration.
    """
    @property
    def duration_label(self):
        if not self.duration:
            return ''
        return self.duration.label



from __future__ import unicode_literals

from ..base import Resource

from .agency import Agency


class Agent(Resource):
    _resource_name = 'agents'
    _is_listable = False

    _as_is_fields = [
        'id', 'href', 'role', 'first_name', 'last_name', 'email',
        'phone_numbers', 'username', 'active',
    ]
    _resource_fields = [('agency', Agency)]



import requests

from . import __title__, __version__


ACCEPTABLE_RESPONSE_STATUS_CODES = (
    requests.codes.ok, requests.codes.created, requests.codes.accepted,
)

JSON_CONTENT_TYPE = 'application/json'


class APIRequestor(object):

    def __init__(self, client, resource, params=None, parent=None):
        self.client = client
        self.resource = resource
        self.params = params
        self.parent = parent

    def _request(self, uri, method, data=None, params=None, additional_headers=None):
        """Make an HTTP request to a target API method with proper headers."""

        assert method in ['GET', 'POST', 'PUT', 'PATCH'], "Only 'GET', 'POST', 'PUT', and 'PATCH' are allowed."

        url = self._get_url(uri)
        headers = self._get_headers(method, additional_headers)
        response = self._make_call(method, url, headers, data, params)
        return response

    def _get_url(self, uri):
        """Return the full URL to make a request to for the given `uri`"""

        # Support supplying a full url
        if '://' in uri:
            url = uri
        else:
            url = self.client.api_root + uri

        # Strip out the proxy from the url. The client only wants to return urls
        # with the API_PROXY, but not actually query on them.
        api_proxy = self.client.api_proxy
        if api_proxy:
            url = url.replace(api_proxy, '')

        return url

    def _get_headers(self, method, additional_headers):
        """Return a dictionary of HTTP headers to set on the request to the API."""

        headers = {
            'User-Agent': '{0}/{1}'.format(__title__, __version__),
            'X-Application-Key': self.client.application_key,
        }

        # gapipy works in JSON. Ensure the receiving API is aware of the type of
        # payload being sent.
        if method in ('POST', 'PUT', 'PATCH'):
            headers['Content-Type'] = JSON_CONTENT_TYPE

        if self.client.api_language:
            headers['Accept-Language'] = self.client.api_language

        if additional_headers:
            headers.update(additional_headers)

        if self.client.api_proxy:
            headers.update({'X-Api-Proxy': self.client.api_proxy})

        return headers

    def _make_call(self, method, url, headers, data, params):
        """Make the actual request to the API, using the given URL, headers,
        data and extra parameters.
        """
        requests_call = getattr(self.client.requestor, method.lower())

        self.client.logger.debug('Making a {0} request to {1}'.format(method, url))

        response = requests_call(url, headers=headers, data=data, params=params)

        if response.status_code in ACCEPTABLE_RESPONSE_STATUS_CODES:
            return response.json()
        else:
            response.reason = response.text
            return response.raise_for_status()

    def get(self, resource_id=None, uri=None, variation_id=None):
        """
        Get a single resource with the given resource_id or uri

        If a resource_id is supplied, a variation_id may also be given -- when
        generating the URI, the variation_id will come after the resource_id,
        separated by a slash.
        """
        if resource_id is None and uri is None:
            raise ValueError(
                'Need to provide at least one of `resource_id` or `uri` as argument')
        if not uri:
            components = ['', self.resource, str(resource_id)]
            if variation_id:
                components.append(str(variation_id))
            uri = '/'.join(components)
        return self._request(uri, 'GET')

    def update(self, resource_id, data, partial=True, uri=None):
        """
        Update a single resource with the given data.

        When `partial` is True, the http method is `PATCH`. Otherwise, `PUT` is
        sent.
        """
        method = 'PATCH' if partial else 'PUT'
        if not uri:
            uri = '/{0}/{1}'.format(self.resource, resource_id)
        return self._request(uri, method, data=data)

    def create(self, data, uri=None):
        """
        Create a single new resource with the given data.
        """
        if not uri:
            uri = '/{0}'.format(self.resource)
        return self._request(uri, 'POST', data=data)

    def list_raw(self, uri=None):
        """Return the raw response for listing resources.

        If `uri` is given, make a GET request to it, otherwise build the uri
        from `self.resource` and `self.parent`. Note that only the first page
        of results will be returned. To get an iterator over all resources,
        use `list`.
        """
        if not uri:
            if self.parent:
                parent_name, parent_id = self.parent
                uri = '/{0}/{1}/{2}'.format(parent_name, parent_id, self.resource)
            else:
                uri = '/{0}'.format(self.resource)

        return self._request(uri, 'GET', params=self.params)

    def list(self, uri=None):
        """Generator for listing resources"""

        response = self.list_raw(uri)
        for result in response['results']:
            yield result

        for link in response.get('links', []):
            if link['rel'] != 'next':
                continue
            for result in self.list(link['href']):
                yield result



from __future__ import unicode_literals

from ..base import Resource

from .transaction import Payment, Refund
from .document import Invoice, Document
from .service import Service


class Booking(Resource):
    _resource_name = 'bookings'
    _is_parent_resource = True

    _as_is_fields = ['id', 'href', 'external_id', 'currency']
    _price_fields = [
        'amount_paid', 'amount_owing', 'commission', 'tax_on_commission',
    ]
    _date_fields = [
        'date_closed', 'date_of_first_travel', 'date_of_last_travel',
        'balance_due_date',
    ]
    _date_time_fields_utc = ['date_created', ]
    _resource_fields = [('agent', 'Agent')]

    @property
    def _resource_collection_fields(self):
        return [
            ('services', Service),
            ('invoices', Invoice),
            ('payments', Payment),
            ('refunds', Refund),
            ('documents', Document),
        ]



from .base import BaseModel, RelatedResourceMixin


class AddOn(BaseModel, RelatedResourceMixin):
    _as_is_fields = ['max_days', 'min_days', 'product']
    _date_fields = ['start_date', 'finish_date', 'halt_booking_date', 'request_space_date']
    _resource_fields = []

    def __repr__(self):
        return '<{0} ({1})>'.format(self.__class__.__name__, self.product.name)

    # product is marked as as-is, so we can turn it into a dynamically-typed resource, based on product['type']
    # Is there a better way?
    def _fill_fields(self, data):
        super(AddOn, self)._fill_fields(data)
        r = {
            'activities': 'Activity',
            'accommodations': 'Accommodation',
            'transports': 'Transport',
            'single_supplements': 'SingleSupplement',
        }.get(self.product['type'])
        if r:
            self._resource_fields += [('product', r)]
            self._set_resource_field('product', self.product)



from __future__ import unicode_literals

from ..base import Resource
from .state import State


class Country(Resource):

    _resource_name = 'countries'
    _is_parent_resource = True

    _as_is_fields = ['id', 'href', 'name']
    _resource_collection_fields = [('states', State)]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from unittest import TestCase

from gapipy.utils import (
    dict_to_model,
    duration_label,
    humanize_amount,
    humanize_price,
    humanize_time,
    location_label,
)


class UtilsTestCase(TestCase):

    def test_humanize_amount(self):
        self.assertEqual(humanize_amount('1'), '1')
        self.assertEqual(humanize_amount('1.5'), '1.50')
        self.assertEqual(humanize_amount('1.500'), '1.50')
        self.assertEqual(humanize_amount('0.50'), '0.50')
        self.assertEqual(humanize_amount('1', force_decimal=True), '1.00')

    def test_humanize_price_free(self):
        FREE = 'Free'
        self.assertEqual(humanize_price(0, None, 'CAD'), FREE)
        self.assertEqual(humanize_price(0.00, None, 'CAD'), FREE)
        self.assertEqual(humanize_price('0', None, 'CAD'), FREE)
        self.assertEqual(humanize_price('0.00', None, 'CAD'), FREE)
        self.assertEqual(humanize_price('0.00', 5.00, 'CAD'), 'Free-5CAD')

    def test_humanize_price_none(self):
        out = humanize_price(None, None, 'CAD')
        self.assertEqual(out, '')

    def test_humanize_price_range(self):
        self.assertEqual(humanize_price(1, 2, 'CAD'), '1-2CAD')
        self.assertEqual(humanize_price(2.5, 3.0, 'CAD'), '2.50-3.00CAD')
        self.assertEqual(humanize_price(2.5000, 3.1000, 'CAD'), '2.50-3.10CAD')
        # As strings
        self.assertEqual(humanize_price('2.5', '3.1', 'CAD'), '2.50-3.10CAD')

    def test_humanize_price_single(self):
        self.assertEqual(humanize_price(1, None, 'CAD'), '1CAD')
        self.assertEqual(humanize_price('1.50', None, 'CAD'), '1.50CAD')

    def test_humanize_time(self):
        self.assertEqual(humanize_time('1.5'), '1h30m')
        self.assertEqual(humanize_time('0.5'), '30m')
        self.assertEqual(humanize_time(1.0), '1h')

    def test_duration_label(self):
        self.assertEqual(duration_label(1.5, 2.5), '1h30m-2h30m')
        self.assertEqual(duration_label(1, None), '1h')
        self.assertEqual(duration_label('3.0', None), '3h')

    def test_location_label(self):
        import collections
        Place = collections.namedtuple('Place', ['name'])
        place_1 = Place(name='Toronto')
        place_2 = Place(name='Montreal')
        self.assertEqual(location_label(place_1, place_2), 'Toronto – Montreal')
        self.assertEqual(location_label(place_1, place_1), 'Toronto')

    def test_dict_to_model(self):
        data = {
            'id': '123',
            'name': {
                'first': 'Foo',
                'last': 'Baz',
                'reverse': {
                    'first': 'Oof',
                    'last': 'Zab',
                },
            },
            'phone_numbers': [
                {
                    'number': '555-555-5555',
                }

            ],
        }

        wrapper = dict_to_model('Profile')
        model = wrapper(data)

        self.assertEqual(str(model), 'Profile')
        self.assertEqual(repr(model), '<Profile>')
        self.assertEqual(model.id, '123')
        self.assertEqual(model.name.first, 'Foo')
        self.assertEqual(model.name.last, 'Baz')
        self.assertEqual(str(model.name), 'Name')

        self.assertEqual(model.name.reverse.first, 'Oof')
        self.assertEqual(model.name.reverse.last, 'Zab')

        self.assertEqual(str(model.phone_numbers[0]), 'Phone Numbers')
        self.assertEqual(model.phone_numbers[0].number, '555-555-5555')



from __future__ import unicode_literals

from ..base import Resource
from .details import DossierDetail, DossierDetailsMixin


class AccommodationDossier(Resource, DossierDetailsMixin):
    _resource_name = 'accommodation_dossiers'

    _as_is_fields = [
        'id', 'href', 'name',
        'website',
        'property_type',
        'address',
        'features',
        'emails',
        'phone_numbers',
        'rooms',
    ]

    _date_time_fields_local = ['date_created', 'date_last_modified']

    _resource_fields = [
        ('location', 'Place'),
    ]

    _model_collection_fields = [
        ('details', DossierDetail),
    ]



from __future__ import unicode_literals

from ...models.base import BaseModel
from ..base import Resource
from .details import DossierDetail


class PlaceDossierImage(BaseModel):
    _as_is_fields = ['id', 'href', 'date_created', 'date_last_modified']


class PlaceDossier(Resource):
    _resource_name = 'place_dossiers'

    _as_is_fields = [
        'id', 'href', 'name'
    ]

    _date_time_fields_local = ['date_created', 'date_last_modified']

    _resource_fields = [
        ('place', 'Place')
    ]

    _model_collection_fields = [
        ('details', DossierDetail),
        ('images', PlaceDossierImage),  # TODO: replace with Image resource
    ]



from __future__ import unicode_literals

from ..base import Resource


class FeatureCategory(Resource):

    _resource_name = 'feature_categories'

    _as_is_fields = [
        'id', 'href', 'code', 'description',
    ]
    _date_time_fields_utc = ['date_created', 'date_last_modified']



from __future__ import unicode_literals

from ...models import Address, AddOn, DepartureRoom, PP2aPrice

from ..base import Product
from .tour_dossier import TourDossier
from .departure_component import DepartureComponent


class Departure(Product):

    _resource_name = 'departures'
    _is_listable = True
    _is_parent_resource = True

    _as_is_fields = [
        'id', 'href', 'availability', 'flags', 'nearest_start_airport',
        'nearest_finish_airport', 'product_line', 'sku', 'requirements',
    ]
    _date_fields = ['start_date', 'finish_date']
    _date_time_fields_utc = ['date_created', 'date_last_modified']
    _date_time_fields_local = ['latest_arrival_time', 'earliest_departure_time']
    _resource_fields = [('tour', 'Tour'), ('tour_dossier', TourDossier)]
    _resource_collection_fields = [
        ('components', DepartureComponent),
    ]
    _model_fields = [('start_address', Address), ('finish_address', Address)]
    _model_collection_fields = [
        ('addons', AddOn),
        ('rooms', DepartureRoom),
        ('lowest_pp2a_prices', PP2aPrice),
    ]
    _deprecated_fields = ['add_ons']



import time
from unittest import TestCase, skipUnless

from gapipy import cache
from gapipy.cache import make_key
from gapipy.client import Client

class CacheTestCase(TestCase):
    def test_make_key(self):
        Client()
        self.assertEquals(make_key('foo', 1), 'foo:1')

    def test_make_key_language(self):
        Client(api_language='de')
        self.assertEquals(make_key('foo', 1), 'foo:1:de')

    def test_make_key_variation_id(self):
        Client()
        self.assertEqual(make_key('foo', 1, 9000), 'foo:1:9000')


class SimpleCacheTestCase(TestCase):

    def test_get_set(self):
        c = cache.SimpleCache()
        c.set('foo', {'xyz': 'bar'})
        data = c.get('foo')
        self.assertEquals(data, {'xyz': 'bar'})

    def test_get_set_resource_id(self):
        c = cache.SimpleCache()
        c.set('foo', {'id': 100, 'xyz': 'bar'})
        data = c.get('foo', 100)
        self.assertEquals(data, {'id': 100, 'xyz': 'bar'})

    def test_get_set_resource_and_variation_id(self):
        c = cache.SimpleCache()
        c.set('foo', {'id': 100, 'xyz': 'bar', 'variation_id': 9000})
        data = c.get('foo', 100, 9000)
        self.assertEquals(data, {'id': 100, 'xyz': 'bar', 'variation_id': 9000})

    def test_get_many_resource_id(self):
        c = cache.SimpleCache()
        c.set('a', {'id': 1, 0: 0})
        c.set('a', {'id': 2, 1: 1})

        self.assertEquals(c.get_many('a', (1, 2)), [
            {'id': 1, 0: 0},
            {'id': 2, 1: 1},
        ])

    def test_set_many(self):
        c = cache.SimpleCache()
        c.set_many('foo', [
            {'id': 1, 0: 0},
            {'id': 2, 1: 1},
            {'id': 3, 2: 4},
        ])
        self.assertEquals(c.get('foo', 2), {'id': 2, 1: 1})
        c.set_many('foo', [{'id': 2, 1: 100}])
        self.assertEquals(c.get('foo', 2), {'id': 2, 1: 100})

    def test_delete(self):
        c = cache.SimpleCache()
        c.set('xyz', {'foo': 'bar'})
        self.assertEquals(c.get('xyz'), {'foo': 'bar'})
        c.delete('xyz')
        self.assertEquals(c.get('xyz'), None)


def _redis_installed():
    """Returns True if and only if the redis module is available"""
    try:
        import redis  # NOQA
        return True
    except ImportError:
        return False


@skipUnless(_redis_installed(), 'redis is not installed')
class RedisCacheTestCase(TestCase):
    def make_cache(self):
        return cache.RedisCache(key_prefix='gapi-python-client-test-case:')

    def tearDown(self):
        self.make_cache().clear()

    def test_get_set(self):
        c = self.make_cache()
        c.set('a', {'id': 1})
        self.assertEquals(c.get('a', 1), {'id': 1})

        c.set('a', {'results': (1, 2, 3)})
        self.assertEquals(c.get('a'), {'results': (1, 2, 3)})

    def test_get_set_resource_and_variation_id(self):
        c = self.make_cache()
        c.set('foo', {'id': 100, 'xyz': 'bar', 'variation_id': 9000})
        data = c.get('foo', 100, 9000)
        self.assertEquals(data, {'id': 100, 'xyz': 'bar', 'variation_id': 9000})

    def test_get_many(self):
        c = self.make_cache()
        c.set('a', {'id': 1})
        c.set('a', {'id': 2})
        self.assertEquals(c.get_many('a', (1, 2)), [
            {'id': 1},
            {'id': 2},
        ])

    def test_set_many(self):
        c = self.make_cache()
        c.set_many('a', [
            {'id': 1, 0: 0},
            {'id': 2, 1: 1},
            {'id': 3, 2: 4},
        ])
        self.assertEquals(c.get('a', 3), {'id': 3, 2: 4})

    def test_expire(self):
        c = self.make_cache()
        c.set('expire_me', {0: 0}, timeout=1)
        time.sleep(2)
        self.assertEquals(c.get('expire_me'), None)

    def test_delete(self):
        c = self.make_cache()
        c.set('delete_me', {0: 0})
        self.assertEquals(c.get('delete_me'), {0: 0})
        c.delete('delete_me')
        self.assertEquals(c.get('delete_me'), None)



from .base import BaseModel

class AssociatedService(BaseModel):
    """
    Represent an associated service. Each service can be associated
    to other services within the same booking.
    """
    _as_is_fields = ['id', 'href']



from __future__ import unicode_literals

from ...models import PriceBand

from ..base import Product


class SingleSupplement(Product):

    _resource_name = 'single_supplements'
    _is_listable = False

    _as_is_fields = ['id', 'href', 'availability', 'name', 'product_line', 'sku', 'type', 'sub_type']
    _date_fields = ['start_date', 'finish_date']
    _date_time_fields_utc = ['date_created', 'date_last_modified']
    _model_collection_fields = [('price_bands', PriceBand)]
    _resource_fields = [('departure', 'Departure')]



from .tour import (  # NOQA
    Accommodation,
    Activity,
    OptionalActivity,
    Departure,
    DepartureComponent,
    Duration,
    Itinerary,
    Promotion,
    SingleSupplement,
    Tour,
    TourCategory,
    TourDossier,
    Transport,
)

from .dossier import (  # NOQA
    AccommodationDossier,
    ActivityDossier,
    PlaceDossier,
    TransportDossier,
    DossierDetail,
    DossierDetailType,
)

from .booking import (  # NOQA
    Agency,
    Agent,
    Booking,
    Customer,
    Invoice,
    DeclinedReason,
    Document,
    Nationality,
    AccommodationService,
    ActivityService,
    DepartureService,
    FeeService,
    FlightService,
    InsuranceService,
    SingleSupplementService,
    TransportService,
    Payment,
    Refund,
)

from .geo import (  # NOQA
    Airport,
    Continent,
    Country,
    Feature,
    FeatureCategory,
    Place,
    State,
    Timezone,
)

from .language import Language  # NOQA


available_public_resources = [
    # Tour
    'Accommodation',
    'Activity',
    'Departure',
    'DepartureComponent',
    'Itinerary',
    'Promotion',
    'SingleSupplement',
    'Tour',
    'TourCategory',
    'TourDossier',
    'Transport',

    # Geographical
    'Airport',
    'Continent',
    'Country',
    'Feature',
    'FeatureCategory',
    'Place',
    'State',
    'Timezone',

    # Dossier
    'AccommodationDossier',
    'ActivityDossier',
    'PlaceDossier',
    'TransportDossier',
    'Language',
]

available_private_resources = [
    'AccommodationService',
    'ActivityService',
    'Agency',
    'Agent',
    'Booking',
    'Customer',
    'DeclinedReason',
    'DepartureService',
    'Document',
    'FeeService',
    'FlightService',
    'InsuranceService',
    'Invoice',
    'Nationality',
    'Payment',
    'Refund',
    'SingleSupplementService',
    'TransportService',
]

available_resources = available_public_resources + available_private_resources



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from ..base import Resource
from ...models.base import BaseModel
from ...utils import DurationLabelMixin, LocationLabelMixin, duration_label


class OptionalActivity(BaseModel):
    _resource_fields = [
        ('activity_dossier', 'ActivityDossier'),
    ]

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.activity_dossier)


class Duration(BaseModel):
    _as_is_fields = [
        'min_hr', 'max_hr',
    ]

    def __repr__(self):
        return '<{}: {},{}>'.format(self.__class__.__name__, self.min_hr, self.max_hr)

    @property
    def label(self):
        return duration_label(self.min_hr, self.max_hr)


class ItineraryComponent(BaseModel, DurationLabelMixin, LocationLabelMixin):
    _as_is_fields = [
        'type',
        'summary', 'description', 'instructions',
        'distance_km', 'start_time', 'end_time', 'time_period', 'is_overnight',
    ]
    _resource_fields = [
        ('start_location', 'Place'),
        ('end_location', 'Place'),
        ('activity_dossier', 'ActivityDossier'),
        ('transport_dossier', 'TransportDossier'),
    ]
    _model_fields = [
        ('duration', 'Duration'),
    ]

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.type)

    @property
    def is_free_time(self):
        return self.type == 'FREE_TIME'

    @property
    def is_activity(self):
        return self.type == 'ACTIVITY'

    @property
    def is_transport(self):
        return self.type == 'TRANSPORT'


class ItineraryDay(BaseModel):
    _as_is_fields = [
        'id', 'day', 'summary', 'description', 'instructions', 'meals',
    ]
    _resource_fields = [
        ('start_location', 'Place'),
        ('end_location', 'Place'),
    ]
    _model_collection_fields = [
        ('components', ItineraryComponent),
        ('optional_activities', OptionalActivity),
    ]

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.day)


class ValidDuringRange(BaseModel):
    _date_fields = ['start_date', 'end_date']

    def __repr__(self):
        return '<{} ({} - {})>'.format(self.__class__.__name__, self.start_date, self.end_date)


class Itinerary(Resource):

    _resource_name = 'itineraries'

    _as_is_fields = [
        'id', 'href', 'name', 'duration', 'meals_included', 'meals_budget',
        'packing_lists', 'images', 'variation_id',
    ]
    _resource_fields = [
        ('start_location', 'Place'),
        ('end_location', 'Place'),
    ]
    _model_collection_fields = [
        ('days', ItineraryDay),
        ('variations', 'Itinerary'),
        ('valid_during_ranges', ValidDuringRange),
    ]



from .base import BaseModel


class Address(BaseModel):
    _as_is_fields = ['city', 'latitude', 'longitude', 'postal_zip', 'street']
    _resource_fields = [
        ('state', 'State'),
        ('country', 'Country')
    ]

    def __repr__(self):
        return '<{0}: {1}, {2}>'.format(
            self.__class__.__name__, self.city, self.country.name)



from functools import wraps
from itertools import islice

from requests import HTTPError

from .request import APIRequestor


def _check_listable(func):

    @wraps(func)
    def inner(query, *args, **kwargs):
        if not (query.resource._is_listable or query.parent):
            raise ValueError(
                'The {0} resource is not listable and/or is only available as a subresource'.format(
                    query.resource.__name__))

        return func(query, *args, **kwargs)

    return inner


class Query(object):

    def __init__(self, client, resource, filters=None, parent=None, raw_data=None):
        self._client = client
        self._raw_data = raw_data or {}
        self.resource = resource
        self.filters = filters or {}
        self.parent = parent

    def _to_dict(self):
        # Used by Resource when converting nested Query objects into
        # serializable types.
        return self._raw_data

    def get(self, resource_id, variation_id=None, cached=True):
        """
        Returns an instance of the query resource with the given `resource_id`
        (and optional `variation_id`) or `None` if the resource with the given
        id does not exist.

        If the resource is available in the cache and `cached` is True, no http
        request will be made. If `cached` is False, a fresh request will be
        made for the resource, which will be re-added to the cache. This is a
        good method to invalidate persistent cache backend after receiving a
        webhook that a resource has changed.
        """
        try:
            data = self.get_resource_data(resource_id,
                variation_id=variation_id, cached=cached)
        except HTTPError as e:
            if e.response.status_code == 404:
                return None
            raise e
        resource_object = self.resource(data)
        self._client._cache.set(self.resource._resource_name, resource_object.to_dict())
        return resource_object

    def get_resource_data(self, resource_id, variation_id=None, cached=True):
        '''
        Returns a dictionary of resource data, which is used to initialize
        a Resource object in the `get` method.
        '''
        resource_data = None
        if cached:
            resource_data = self._client._cache.get(
                self.resource._resource_name,
                resource_id,
                variation_id,
            )
            if resource_data is not None:
                return resource_data

        # Cache miss; get fresh data from the backend.
        requestor = APIRequestor(self._client, self.resource._resource_name)
        return requestor.get(resource_id, variation_id=variation_id)

    def purge_cached(self, resource_id, variation_id=None):
        return self._client._cache.delete(
            self.resource._resource_name,
            resource_id,
            variation_id,
        )

    def is_cached(self, resource_id, variation_id=None):
        return self._client._cache.is_cached(
            self.resource._resource_name,
            resource_id,
            variation_id,
        )

    @_check_listable
    def all(self, limit=None):
        """Generator of instances of the query resource. If limit is set to a
        positive integer `n`, only return the first `n` results.
        """

        requestor = APIRequestor(
            self._client,
            self.resource._resource_name,
            params=self.filters,
            parent=self.parent
        )
        generator = requestor.list()

        if limit:
            if isinstance(limit, int) and limit > 0:
                generator = islice(generator, limit)
            else:
                raise ValueError('`limit` must be a positive integer')

        for result in generator:
            yield self.resource(result, stub=True)

    def filter(self, **kwargs):
        """Add filter arguments to the query.

        For example, if `query` is a Query for the TourDossier ressource, then
        `query.filter(name='Amazing Adventure')` will return a query containing
        only dossiers whose names contain 'Amazing Adventure'.
        """

        self.filters.update(kwargs)
        return self

    @_check_listable
    def count(self):
        """Returns the number of element in the query."""

        requestor = APIRequestor(
            self._client,
            self.resource._resource_name,
            params=self.filters,
            parent=self.parent
        )
        response = requestor.list_raw()
        return response.get('count')

    def create(self, data_dict):
        """Create an instance of the query resource using the given data"""
        return self.resource.create(self._client, data_dict)

    def __iter__(self):
        """Provided as a convenience so that Query objects can be iterated
        without calling `all`.

            i.e.  `for dossier in dossiers.filter(name='Peru')`
                instead of `for dossier in dossiers.filter(name='Peru').all()`
        """
        return self.all()



from __future__ import unicode_literals

from ...models import Address, SeasonalPriceBand

from ..base import Product


class Transport(Product):

    _resource_name = 'transports'
    _is_listable = False

    _as_is_fields = [
        'id', 'href', 'availability', 'name', 'product_line', 'sku', 'type', 'sub_type'
    ]
    _date_time_fields_utc = ['date_created', 'date_last_modified']
    _model_fields = [('start_address', Address), ('finish_address', Address)]
    _model_collection_fields = [('price_bands', SeasonalPriceBand)]



# flake8: NOQA
from .airport import Airport
from .continent import Continent
from .country import Country
from .feature import Feature
from .feature_category import FeatureCategory
from .place import Place
from .state import State
from .timezone import Timezone



from __future__ import unicode_literals

from ...query import Query
from ...utils import get_resource_class_from_class_name

from ..base import Resource
from ...models import AdvertisedDeparture


BRIEF_ITINERARY_TYPE = 'SUMMARY'
DETAILED_ITINERARY_TYPE = 'DETAILED'
MAP_IMAGE_TYPE = 'MAP'
BANNER_IMAGE_TYPE = 'BANNER'


class TourDossier(Resource):

    _resource_name = 'tour_dossiers'

    _as_is_fields = [
        'id', 'href', 'categories', 'description', 'details', 'product_line',
        'geography', 'images', 'itineraries', 'name', 'site_links',
    ]
    _date_fields = ['departures_start_date', 'departures_end_date']
    _resource_fields = [('tour', 'Tour')]
    _resource_collection_fields = [
        ('departures', 'Departure'),
        ('structured_itineraries', 'Itinerary'),
    ]
    _model_collection_fields = [
        ('advertised_departures', AdvertisedDeparture),
    ]

    def _set_resource_collection_field(self, field, value):
        """Overridden to ensure that the `departures` query has the right
        parent resource (i.e. the tour and not the tour dossier), and to handle
        the special case of structured itineraries being an associated resource
        that is not a child resource.
        """

        from .itinerary import Itinerary

        if field == 'departures':
            resource_cls = get_resource_class_from_class_name('Departure')

            # Tour dossiers always have the same id as the corresponding tour
            parent = ('tours', self.id)

            setattr(self, 'departures', Query(self._client, resource_cls, parent=parent, raw_data=value))

        elif field == 'structured_itineraries':
            itineraries = [Itinerary(i, stub=True) for i in value]
            setattr(self, 'structured_itineraries', itineraries)

        else:
            return super(TourDossier, self)._set_resource_collection_field(field, value)

    def _get_itinerary(self, itinerary_type):
        for itin in self.itineraries:
            if itin['type'] == itinerary_type:
                return [dict(label=i['label'], body=i['body']) for i in itin['days']]

    def get_brief_itinerary(self):
        return self._get_itinerary(BRIEF_ITINERARY_TYPE)

    def get_detailed_itinerary(self):
        return self._get_itinerary(DETAILED_ITINERARY_TYPE)

    def _get_image_url(self, image_type):
        for image in self.images:
            if image['type'] == image_type:
                return image['image_href']

    def get_map_url(self):
        return self._get_image_url(MAP_IMAGE_TYPE)

    def get_banner_url(self):
        return self._get_image_url(BANNER_IMAGE_TYPE)

    def get_visited_countries(self):
        return [country['name'] for country in self.geography['visited_countries']]

    def get_trip_detail(self, label):
        for detail in self.details:
            if detail['detail_type']['label'] == label:
                return detail['body']



from ..resources.base import Resource
from ..utils import get_resource_class_from_resource_name


class PricePromotion(Resource):
    """
    The `promotions` referenced in a `Price` are not valid, as they include an
    `amount` field that is not part of the Promotion object.

    Thus, this intermediatary model must be used to represent a pseudo-Resource.
    """
    def __init__(self, data, **kwargs):
        self._resource_name = 'promotions'

        klass = get_resource_class_from_resource_name('promotions')
        for k, v in klass.__dict__.items():
            if 'fields' in k and isinstance(v, list):
                setattr(self, k, getattr(klass, k))

        super(PricePromotion, self).__init__(data, **kwargs)

        # Add the "fake" amount field.
        setattr(self, 'amount', data['amount'])



from __future__ import unicode_literals

from ...models import Address
from ...models import AgencyDocument

from ..base import Resource
from ..tour import Promotion


class Agency(Resource):
    _resource_name = 'agencies'
    _is_listable = False
    _is_parent_resource = True

    _as_is_fields = ['id', 'href', 'name', 'booking_currencies', 'latitude', 'longitude']
    _date_time_fields_local = ['date_created']
    _model_fields = [('address', Address)]
    _model_collection_fields = [('documents', AgencyDocument)]
    _resource_collection_fields = [
        ('bookings', 'Booking'),
        ('agents', 'Agent'),
        ('promotions', Promotion),
    ]



from __future__ import unicode_literals

from ...utils import get_resource_class_from_resource_name
from ..base import Resource, Product

class PromotionProduct(Resource):
    """
    The `products` referenced in a Promotion object are not valid resources (due
    to the existence of `type`) so they must be wrapped in this model.

    When the resource is fixed, the hacks here can simply be replaced with
    `gapipy.models.base.RelatedResourceMixin`
    """
    def __init__(self, data, **kwargs):
        # Fetch the resource class using the `type`, and then derive field
        # attributes from that class.
        klass = get_resource_class_from_resource_name(data['type'])
        for k, v in klass.__dict__.items():
            if 'fields' in k and isinstance(v, list):
                setattr(self, k, getattr(klass, k))

        self._resource_name = data['type']
        self._as_is_fields = self._as_is_fields + ['type', 'sub_type']

        super(PromotionProduct, self).__init__(data, **kwargs)

class Promotion(Resource):

    _resource_name = 'promotions'

    _as_is_fields = [
        'id', 'href', 'commission_rate', 'currencies', 'flags',
        'min_travellers', 'name', 'promotion_code', 'room_codes',
        'terms_and_conditions',
    ]

    _date_fields = [
        'product_start_date', 'product_finish_date',
        'sale_start_date', 'sale_finish_date'
    ]
    _price_fields = ['discount_amount', 'discount_percent']
    _model_collection_fields = [('products', Product)]

    def _set_model_collection_field(self, field, data):
        # Only products needs special treatment.
        if field != 'products':
            return super(Promotion, self)._set_model_collection_field(field, data)

        product_stubs = []
        # Each product can be a different resource, so derive the resource from
        # the "type" within the stubbed object.
        for product in data:
            stub = PromotionProduct(product, stub=True)
            product_stubs.append(stub)
        setattr(self, field, product_stubs)



import json
import unittest

from mock import patch
from requests import HTTPError, Response

from gapipy.client import Client
from gapipy.query import Query
from gapipy.resources import Accommodation, Departure, Tour, TourDossier
from gapipy.resources.base import Resource

from .fixtures import (
    PPP_TOUR_DATA, TOUR_DOSSIER_LIST_DATA, DUMMY_DEPARTURE, DUMMY_PROMOTION,
)


class QueryTestCase(unittest.TestCase):

    def setUp(self):
        self.client = Client()
        self.cache = self.client._cache
        self.cache.clear()

    @patch('gapipy.request.APIRequestor._request', return_value=PPP_TOUR_DATA)
    def test_get_instance_by_id(self, mock_request):
        query = Query(self.client, Tour)
        t = query.get(1234)
        self.assertIsInstance(t, Tour)

    @patch('gapipy.request.APIRequestor._request')
    def test_get_instance_with_non_existing_id(self, mock_request):
        response = Response()
        response.status_code = 404
        http_error = HTTPError(response=response)
        mock_request.side_effect = http_error

        query = Query(self.client, Tour)
        t = query.get(1234)
        self.assertIsNone(t)

    @patch('gapipy.request.APIRequestor._request')
    def test_get_instance_by_id_with_non_404_error(self, mock_request):
        response = Response()
        response.status_code = 401
        http_error = HTTPError(response=response)
        mock_request.side_effect = http_error

        query = Query(self.client, Tour)
        with self.assertRaises(HTTPError) as cm:
            query.get(1234)

        self.assertEqual(cm.exception.response.status_code, 401)

    @patch('gapipy.request.APIRequestor._request')
    def test_filtered_query(self, mock_request):
        query = Query(self.client, Tour).filter(tour_dossier_code='PPP')
        list(query)  # force query evaluation
        mock_request.assert_called_once_with(
            '/tours', 'GET', params={'tour_dossier_code': 'PPP'})

        # Check that filters can be chained
        list(query.filter(departures_start_date='2014-01-01'))
        mock_request.assert_called_with(
            '/tours', 'GET', params={
                'tour_dossier_code': 'PPP',
                'departures_start_date': '2014-01-01'
            })

    def test_listing_non_listable_resource_fails(self):
        message = 'The Accommodation resource is not listable and/or is only available as a subresource'
        with self.assertRaisesRegexp(ValueError, message):
            Query(self.client, Accommodation).all()
        with self.assertRaisesRegexp(ValueError, message):
            Query(self.client, Accommodation).count()

    @patch('gapipy.request.APIRequestor._request', return_value=DUMMY_PROMOTION)
    def test_can_retrieve_single_non_listable_resource(self, mock_request):
        Query(self.client, Accommodation).get(1234)
        mock_request.assert_called_once_with(
            '/accommodations/1234', 'GET')

    @patch('gapipy.request.APIRequestor._request', return_value=DUMMY_DEPARTURE)
    def test_can_retrieve_single_subresource_without_parent(self, mock_request):
        Query(self.client, Departure).get(1234)
        mock_request.assert_called_once_with(
            '/departures/1234', 'GET')

    @patch('gapipy.request.APIRequestor._request', return_value=TOUR_DOSSIER_LIST_DATA)
    def test_count(self, mock_request):
        query = Query(self.client, TourDossier)
        count = query.count()
        self.assertIsInstance(count, int)
        self.assertEqual(count, 3)

    @patch('gapipy.request.APIRequestor._request', return_value=TOUR_DOSSIER_LIST_DATA)
    def test_fetch_all(self, mock_request):

        query = Query(self.client, TourDossier).all()
        self.assertFalse(mock_request.called)

        dossiers = list(query)
        self.assertTrue(mock_request.called)

        self.assertEqual(len(dossiers), 3)
        for dossier in dossiers:
            self.assertIsInstance(dossier, TourDossier)

        mock_request.assert_called_once_with(
            '/tour_dossiers', 'GET', params={})

    @patch('gapipy.request.APIRequestor._request', return_value=TOUR_DOSSIER_LIST_DATA)
    def test_fetch_all_with_limit(self, mock_request):

        query = Query(self.client, TourDossier).all(limit=2)
        dossiers = list(query)

        self.assertEqual(len(dossiers), 2)
        for dossier in dossiers:
            self.assertIsInstance(dossier, TourDossier)

        mock_request.assert_called_once_with(
            '/tour_dossiers', 'GET', params={})

    def test_fetch_all_with_wrong_argument_for_limit(self):
        message = '`limit` must be a positive integer'
        with self.assertRaisesRegexp(ValueError, message):
            query = Query(self.client, Tour).all(limit=-1)
            list(query)  # force the query to evaluate


class QueryCacheTestCase(unittest.TestCase):
    def setUp(self):
        self.client = Client(cache_backend='gapipy.cache.SimpleCache')
        self.cache = self.client._cache
        self.cache.clear()

    @patch('gapipy.request.APIRequestor._request', return_value=PPP_TOUR_DATA)
    def test_resources_are_cached(self, mock_request):
        query = Query(self.client, Tour)

        self.assertEqual(self.cache.count(), 0)

        query.get(21346)
        self.assertEqual(self.cache.count(), 1)
        self.assertEqual(len(mock_request.mock_calls), 1)

        query.get(21346)
        self.assertEqual(len(mock_request.mock_calls), 1)


class MockResource(Resource):
    _as_is_fields = ['id', 'first_name', 'last_name']
    _resource_name = 'mocks'


@patch('gapipy.request.APIRequestor._request')
class UpdateCreateResourceTestCase(unittest.TestCase):
    def test_object_accessor(self, mock_request):
        data = {'first_name': 'Jon', 'last_name': 'Ive', 'id': None}
        r = MockResource(data)
        self.assertEquals(r.first_name, data['first_name'])

    def test_object_attr_modify(self, mock_request):
        data = {'first_name': 'Jon', 'last_name': 'Ive', 'id': None}
        r = MockResource(data)
        r.first_name = 'Jonathan'
        self.assertEquals(r.first_name, 'Jonathan')

    def test_create_object(self, mock_request):
        data = {
            'first_name': 'Jon',
            'last_name': 'Ive',
            'id': None,
        }
        r = MockResource(data)
        r.save()
        mock_request.assert_called_once_with(
            '/mocks', 'POST', data=json.dumps(data))

    def test_update_object(self, mock_request):
        data = {
            'first_name': 'Jon',
            'last_name': 'Ive',
            'id': 1,
        }
        r = MockResource(data)

        r.first_name = 'Jonathan'
        r.save()
        data['first_name'] = 'Jonathan'

        mock_request.assert_called_once_with(
            '/mocks/1', 'PUT', data=json.dumps(data))

    def test_partial_update_object(self, mock_request):
        data = {
            'first_name': 'Jon',
            'last_name': 'Ive',
            'id': 1,
        }
        mock_request.return_value = data
        r = MockResource(data)

        r.first_name = 'Jonathan'
        r_data = {
            'first_name': 'Jonathan',
            'last_name': 'Ive',
            'id': 1
        }
        mock_request.return_value = r_data
        r.save(partial=True)
        changed = {'first_name': 'Jonathan'}
        mock_request.assert_called_once_with(
            '/mocks/1', 'PATCH', data=json.dumps(changed))
        self.assertEquals(r.to_dict(), r_data)

        r.last_name = 'Ivey'
        r_data = {
            'first_name': 'Jonathan',
            'last_name': 'Ivey',
            'id': 1,
        }
        mock_request.return_value = r_data
        r.save(partial=True)
        changed = {'last_name': 'Ivey'}
        mock_request.assert_called_with(
            '/mocks/1', 'PATCH', data=json.dumps(changed))
        self.assertEquals(r.to_dict(), r_data)



from __future__ import unicode_literals

from ..base import Resource
from .departure import Departure
from .tour_dossier import TourDossier


class Tour(Resource):

    _resource_name = 'tours'
    _is_parent_resource = True

    _as_is_fields = ['id', 'href', 'product_line']
    _date_fields = ['departures_start_date', 'departures_end_date']
    _resource_fields = [('tour_dossier', TourDossier)]
    _resource_collection_fields = [('departures', Departure)]

    def get_brief_itinerary(self):
        return self.tour_dossier.get_brief_itinerary()

    def get_detailed_itinerary(self):
        return self.tour_dossier.get_detailed_itinerary()

    def get_map_url(self):
        return self.tour_dossier.get_map_url()

    def get_banner_url(self):
        return self.tour_dossier.get_banner_url()

    def get_visited_countries(self):
        return self.tour_dossier.get_visited_countries()

    def get_trip_detail(self, label):
        return self.tour_dossier.get_trip_detail(label)



from __future__ import unicode_literals

from ..base import Resource


class Customer(Resource):
    _resource_name = 'customers'
    _is_listable = False
    _is_parent_resource = True

    _as_is_fields = [
        'id', 'href', 'place_of_birth', 'meal_preference', 'meal_notes',
        'emergency_contacts', 'medical_notes', 'phone_numbers',
        'account_email', 'name', 'passport', 'address', 'nationality',
    ]

    _date_fields = ['date_of_birth', ]

    @property
    def _resource_collection_fields(self):
        from .booking import Booking
        return [
            ('bookings', Booking),
        ]



import json
import logging
from gapipy.request import APIRequestor
from gapipy.models.base import BaseModel


logger = logging.getLogger(__name__)


class Resource(BaseModel):

    _resource_name = None
    _is_parent_resource = False
    _is_listable = True  # True if resource can be listed/queried (i.e /{resource_name} is an endpoint)

    def __init__(self, data, stub=False, client=None):
        self.is_stub = stub
        super(Resource, self).__init__(data, client)

    def fetch(self):
        logger.info('Fetching %s/%s', self._resource_name, self.id)
        self.is_stub = False

        # Fetch the resource using the client bound on it, which handles cache get/set.
        resource_obj = getattr(self._client, self._resource_name).get(
            self.id,
            variation_id=getattr(self, 'variation_id', None))
        if resource_obj:
            self._fill_fields(resource_obj._raw_data)

        return self

    @classmethod
    def create(cls, client, data_dict):
        request = APIRequestor(client, cls._resource_name)
        response = request.create(json.dumps(data_dict))
        return cls(response)

    def __getattr__(self, name):
        # If we try to access a field that's allowed, and this resource is a
        # stub, we need to fetch it.
        if name in self._allowed_fields() and self.is_stub:
            self.fetch()
            return self.__getattribute__(name)

        raise AttributeError("%r has no field %r available" % (type(self).__name__, name))

    def __repr__(self):
        if self.is_stub:
            return '<{}: {} (stub)>'.format(self.__class__.__name__, self.id)
        else:
            return '<{}: {}>'.format(self.__class__.__name__, self.id)

    def __eq__(self, other):
        # Same resource name and ID determine equality
        return (
            self._resource_name == getattr(other, '_resource_name', None)
            and self.id == other.id
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_json(self):
        return json.dumps(self.to_dict())

    def _update(self, partial=False):
        request = APIRequestor(self._client, self._resource_name)

        data = self.to_dict()
        if partial:
            data = {k: v for k, v in data.iteritems() if self._raw_data.get(k) != v}

        return request.update(self.id, json.dumps(data), partial=partial)

    def _create(self):
        request = APIRequestor(self._client, self._resource_name)
        return request.create(self.to_json())

    def save(self, partial=False):
        if getattr(self, 'id', None) is not None:
            result = self._update(partial=partial)
        else:
            result = self._create()
        self._fill_fields(result)
        return self


class Product(Resource):
    """This is used to mark those resources that can be attached to a Promotion"""
    pass



from .base import BaseModel


class IncompleteRequirement(BaseModel):
    _as_is_fields = ['type', 'name', 'code', 'message', 'flags', 'details']

    @property
    def _resource_fields(self):
        from gapipy.resources import Customer
        return [
            ('customer', Customer),
        ]



from __future__ import unicode_literals

from ...models.base import BaseModel


class DossierDetailType(BaseModel):
    _as_is_fields = ['code', 'label', 'description']

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.label)


class DossierDetail(BaseModel):
    _as_is_fields = ['body']

    _model_fields = [
        ('detail_type', DossierDetailType),
    ]

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.detail_type)


class DossierDetailsMixin(object):
    """
    Helpers for accessing specific details by type (`code`)
    """
    def _get_detail(self, code, default=None):
        """
        Returns the first detail object with for the given detail_type `code`
        """
        details = [d for d in self.details if d.detail_type.code == code]
        return details[0] if details else default

    def _get_detail_body(self, code):
        """
        Returns the body of a given detail_type's `code`
        """
        detail = self._get_detail(code, None)
        if detail:
            return detail.body

    @property
    def summary(self):
        return self._get_detail_body('COMMON__SUMMARY')

    @property
    def description(self):
        return self._get_detail_body('COMMON__DESCRIPTION')



import logging
import os
import re
from importlib import import_module

from .utils import get_available_resource_classes


current_client = None

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())

default_config = {
    'application_key': os.environ.get('GAPI_APPLICATION_KEY'),
    'api_root': os.environ.get('GAPI_API_ROOT', 'https://rest.gadventures.com'),
    'api_proxy': os.environ.get('GAPI_API_PROXY', ''),
    'api_language': os.environ.get('GAPI_LANGUAGE'),
    'cache_backend': os.environ.get('GAPI_CACHE_BACKEND', 'gapipy.cache.NullCache'),
    'cache_options': {'threshold': 500, 'default_timeout': 3600},
    'debug': os.environ.get('GAPI_CLIENT_DEBUG', False),
    'connection_pool_options': {
        'enable': os.environ.get('GAPI_CLIENT_CONNECTION_POOL_ENABLE', False),
        'block': os.environ.get('GAPI_CLIENT_CONNECTION_POOL_BLOCK', False),
        'number': os.environ.get('GAPI_CLIENT_CONNECTION_POOL_NUMBER', 10),
        'maxsize': os.environ.get('GAPI_CLIENT_CONNECTION_POOL_MAXSIZE', 10),
    },
}

def _get_protocol_prefix(api_root):
    """
    Returns the protocol plus "://" of api_root.

    This is likely going to be "https://".
    """
    match = re.search(r'^[^:/]*://', api_root)
    return match.group(0) if match else ''

def get_config(config, name):
    return config.get(name, default_config[name])


class Client(object):

    def __init__(self, **config):

        global current_client
        current_client = self

        self.application_key = get_config(config, 'application_key')
        self.api_root = get_config(config, 'api_root')
        self.api_proxy = get_config(config, 'api_proxy')
        self.api_language = get_config(config, 'api_language')
        self.cache_backend = get_config(config, 'cache_backend')

        # begin with default connection pool options and overwrite any that the
        # client has specified
        self.connection_pool_options = default_config['connection_pool_options']
        self.connection_pool_options.update(get_config(config, 'connection_pool_options'))


        log_level = 'DEBUG' if get_config(config, 'debug') else 'ERROR'
        self.logger = logger
        self.logger.setLevel(log_level)

        self._set_cache_instance(get_config(config, 'cache_options'))
        self._set_requestor(self.connection_pool_options)

        # Prevent install issues where setup.py digs down the path and
        # eventually fails on a missing requests requirement by importing Query
        # only where it's needed.
        from .query import Query
        for resource in get_available_resource_classes():
            setattr(self, resource._resource_name, Query(self, resource))

    def _set_cache_instance(self, cache_options):
        cache_backend = self.cache_backend
        module_name, class_name = cache_backend.rsplit('.', 1)
        module = import_module(module_name)
        cache = getattr(module, class_name)(**cache_options)
        self._cache = cache

    def _set_requestor(self, pool_options):
        """
        Set the requestor based on connection pooling options.

        If connection pooling is disabled, just set `requests`. If connection
        pooling is enabled, set up a `requests.Session`.
        """
        # We had been importing this at the top of the module, but that seemed
        # to break some CI environments
        import requests

        if not pool_options['enable']:
            self._requestor = requests
            return

        session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(
            pool_block=pool_options['block'],
            pool_connections=pool_options['number'],
            pool_maxsize=pool_options['maxsize'],
        )
        logger.info(
            'Created connection pool (block={}, number={}, maxsize={})'.format(
                pool_options['block'],
                pool_options['number'],
                pool_options['maxsize']))

        prefix = _get_protocol_prefix(self.api_root)
        if prefix:
            session.mount(prefix, adapter)
            logger.info('Mounted connection pool for "{}"'.format(prefix))
        else:
            session.mount('http://', adapter)
            session.mount('https://', adapter)
            logger.info(
                'Could not find protocol prefix in API root, mounted '
                'connection pool on both http and https.')

        self._requestor = session

    @property
    def requestor(self):
        """
        Return a requestor, an object we'll use to make HTTP requests.

        This is either going to be `requests` (if connection pooling is
        disabled), or a `requests.Session` (if connection pooling is enabled), or
        an AttributeError (if `__init__` has not happened yet).
        """
        return self._requestor

    def query(self, resource_name):
        try:
            return getattr(self, resource_name)
        except AttributeError:
            raise AttributeError("No resource named %s is defined." % resource_name)

    def build(self, resource_name, data_dict, **kwargs):
        try:
            resource_cls = getattr(self, resource_name).resource
            return resource_cls(data_dict, **kwargs)
        except AttributeError:
            raise AttributeError("No resource named %s is defined." % resource_name)

    def create(self, resource_name, data_dict):
        """
        Create an instance of the specified resource with `data_dict`
        """
        try:
            resource_cls = getattr(self, resource_name).resource
        except AttributeError:
            raise AttributeError("No resource named %s is defined." % resource_name)

        return resource_cls.create(self, data_dict)



from __future__ import unicode_literals

from ...models.base import BaseModel
from ...utils import get_resource_class_from_resource_name
from ..base import Resource

class AssociatedDossier(BaseModel):
    _as_is_fields = ["type"]
    _model_fields = [
        ("dossier", object),
    ]

    def _set_model_field(self, field, value):
        resource_cls = get_resource_class_from_resource_name(self.type)
        stub = resource_cls(value, stub=True)
        setattr(self, field, stub)

class DepartureComponent(Resource):
    _resource_name = 'departure_components'
    _is_listable = True
    _is_parent_resource = False

    _as_is_fields = [
        "id", "href", "name", "code", "type", "flags",
    ]

    _date_fields = ["start_date"]
    _date_time_fields_local = ["date_created", "date_last_modified"]

    _model_fields = [
        ("associated_dossier", AssociatedDossier),
    ]



from __future__ import unicode_literals

from ..base import Resource
from .details import DossierDetail, DossierDetailsMixin
from ...utils import humanize_price, LocationLabelMixin, DurationLabelMixin


class ActivityDossier(Resource, DossierDetailsMixin, DurationLabelMixin, LocationLabelMixin):
    _resource_name = 'activity_dossiers'

    _as_is_fields = [
        'id', 'href',
        'name',
        'duration_min', 'duration_max',
        'price_per_person_min', 'price_per_person_max',
        'price_per_group_min', 'price_per_group_max',
        'currency',
        'dossier_segment',
        'distance_min', 'distance_max',
    ]

    _date_time_fields_local = ['date_created', 'date_last_modified']

    _resource_fields = [
        ('start_location', 'Place'),
        ('end_location', 'Place'),
    ]

    _model_collection_fields = [
        ('details', DossierDetail),
    ]

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.name)

    @property
    def duration(self):
        """
        Simulate the Duration model in `ItineraryComponent` objects,
        so we can reuse the methods therein until this models inconsistency
        can be rectified (currently, this model has two duration fields
        instead of a duration map/obj).
        """
        if not self.duration_min or not self.duration_max:
            return None

        from gapipy.resources import Duration
        duration = Duration(data=dict(
            min_hr=self.duration_min,
            max_hr=self.duration_max)
        )
        return duration

    @property
    def price_per_person_label(self):
        return humanize_price(
            self.price_per_person_min,
            self.price_per_person_max,
            self.currency,
        )

    @property
    def price_per_group_label(self):
        return humanize_price(
            self.price_per_group_min,
            self.price_per_group_max,
            self.currency,
        )

    @property
    def operational_notes(self):
        return self._get_detail_body('ACTIVITY__OPERATIONAL_NOTES')



# flake8: NOQA
from .details import DossierDetail, DossierDetailType, DossierDetailsMixin
from .accommodation_dossier import AccommodationDossier
from .activity_dossier import ActivityDossier
from .place_dossier import PlaceDossier
from .transport_dossier import TransportDossier



import collections
from functools import partial
from time import time

from gapipy import client as client_module

try:
    import cPickle as pickle
except ImportError:
    import pickle


def make_key(resource_name, resource_id=None, variation_id=None):
    if not resource_id:
        return resource_name

    current_client = client_module.current_client

    parts = [resource_name, str(resource_id)]
    if variation_id:
        parts.append(str(variation_id))

    if current_client.api_language:
        parts.append(current_client.api_language)
    return ':'.join(parts)


def update(d, u):
    for k, v in u.iteritems():
        if isinstance(v, collections.Mapping):
            r = update(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


def _items(mappingorseq):
    """Wrapper for efficient iteraiton over mappings represeted by dicts or
    sequeunces::

        >>> for k, v in _items([i, i*i] for i in xrange(5)):
        ...     assert k*k == v

        >>> for k, v in _items({i: i*i} for i in xrange(5)):
        ...     assert k*k == v
    """
    return mappingorseq.iteritems() if hasattr(mappingorseq, 'iteritems') \
        else mappingorseq


class BaseCache(object):
    """Base class for the cache system. All the cache systems will implemtent
    this API or a superset of it.

    :param default_timeout: the default timeout that is used if no timeout is
                            specified on :meth:`set`.
    """

    def __init__(self, default_timeout=300, **kwargs):
        self.default_timeout = default_timeout

    def get(self, resource_name, resource_id=None, variation_id=None):
        return None

    def set(self, resource_name, data_dict):
        pass

    def set_many(self, resource_name, sequence, timeout=None):
        for data_dict in sequence:
            self.set(resource_name, data_dict, timeout)

    def get_many(self, resource_name, ids):
        func = partial(self.get, resource_name)
        return map(func, ids)

    def delete(self, resource_name, resource_id=None):
        pass

    def clear(self):
        pass

    def count(self):
        raise NotImplementedError

    def is_cached(self, resource_name, resource_id):
        return False


class NullCache(BaseCache):
    """
    A cache that doesn't cache.
    """


class SimpleCache(BaseCache):
    """Simply memory cache for single process environments. This class exists
    mainly for a development server and is not 100% thread safe.

    :param threshold: the maximum number of items the cache stores before
                      it starts evicting keys.
    """
    def __init__(self, threshold=500, default_timeout=300, **kwargs):
        BaseCache.__init__(self, default_timeout)
        self._cache = {}
        self._threshold = threshold

    def _prune(self):
        if len(self._cache) > self._threshold:
            now = time()
            # Prune expired keys, or every few keys.
            for idx, (key, (expires, _)) in enumerate(self._cache.items()):
                if expires <= now or idx % 3 == 0:
                    self._cache.pop(key, None)

    def get(self, resource_name, resource_id=None, variation_id=None):
        key = make_key(resource_name, resource_id, variation_id)
        expires, value = self._cache.get(key, (0, None))
        if expires > time():
            return pickle.loads(value)

    def set(self, resource_name, data_dict, timeout=None):
        key = make_key(resource_name,
            data_dict.get('id'), data_dict.get('variation_id'))
        if timeout is None:
            timeout = self.default_timeout
        self._prune()

        self._cache[key] = (time() + timeout, pickle.dumps(data_dict,
                            pickle.HIGHEST_PROTOCOL))

    def delete(self, resource_name, resource_id=None, variation_id=None):
        key = make_key(resource_name, resource_id, variation_id)
        return self._cache.pop(key, None)

    def clear(self):
        self._cache.clear()

    def count(self):
        return len(self._cache)

    def is_cached(self, resource_name, resource_id, variation_id=None):
        key = make_key(resource_name, resource_id, variation_id)
        return key in self._cache


class RedisCache(BaseCache):
    """Uses the Redis key-value store as a cache backend.
    """
    def __init__(self, host='localhost', port=6379, password=None,
                 db=0, default_timeout=300, key_prefix=None, **kwargs):
        BaseCache.__init__(self, default_timeout)
        self.key_prefix = key_prefix or ''
        try:
            import redis
        except ImportError:
            raise RuntimeError('no redis module found')
        self._client = redis.Redis(host=host, port=port, password=password, db=db)

    def load_object(self, value):
        """The reversal of `dump_object`. This might be called with None.
        """
        if value is None:
            return None
        return pickle.loads(value)

    def dump_object(self, value):
        return pickle.dumps(value)

    def get(self, resource_name, resource_id=None, variation_id=None):
        key = make_key(resource_name, resource_id, variation_id)
        return self.load_object(self._client.get(self.key_prefix + key))

    def set(self, resource_name, data_dict, timeout=None):
        key = make_key(resource_name,
            data_dict.get('id', None), data_dict.get('variation_id'))
        if timeout is None:
            timeout = self.default_timeout
        data = self.dump_object(data_dict)
        return self._client.setex(self.key_prefix + key, data, timeout)

    def delete(self, resource_name, resource_id=None, variation_id=None):
        key = make_key(resource_name, resource_id, variation_id)
        return self._client.delete(self.key_prefix + key)

    def clear(self):
        cache_keys = self._client.keys('{}*'.format(self.key_prefix))
        map(self._client.delete, cache_keys)

    def info(self):
        return self._client.info()

    def is_cached(self, resource_name, resource_id, variation_id=None):
        key = make_key(resource_name, resource_id, variation_id)
        return self._client.exists(self.key_prefix + key)



import datetime
from unittest import TestCase

from mock import patch

from gapipy.query import Query
from gapipy.models import DATE_FORMAT, AccommodationRoom
from gapipy.resources import Tour, TourDossier, Promotion
from gapipy.resources.base import Resource

from .fixtures import PPP_TOUR_DATA, PPP_DOSSIER_DATA


class ResourceTestCase(TestCase):

    def test_to_dict(self):
        t = Tour(PPP_TOUR_DATA)
        self.assertEquals(t.to_dict(), PPP_TOUR_DATA)

    def test_to_dict_datetimes(self):
        class DatetimeResource(Resource):
            _as_is_fields = ['id']
            _date_fields = ['date_field']
            _date_time_fields_utc = ['date_field_utc']

        resource = DatetimeResource({
            'id': 1,
            'date_field': '2013-02-18',
            'date_field_utc': '2013-02-18T18:17:20Z',
        })
        data = resource.to_dict()
        self.assertEquals(data['date_field'], '2013-02-18')
        self.assertEquals(data['date_field_utc'], '2013-02-18T18:17:20Z')

    @patch('gapipy.request.APIRequestor._request', return_value=PPP_DOSSIER_DATA)
    def test_instantiate_from_raw_data(self, mock_request):
        t = Tour(PPP_TOUR_DATA)
        self.assertIsInstance(t, Tour)
        self.assertEqual(t.product_line, 'PPP')
        self.assertIsInstance(t.departures_start_date, datetime.date)
        self.assertIsInstance(t.tour_dossier, TourDossier)
        self.assertIsInstance(t.departures, Query)

    @patch('gapipy.request.APIRequestor._request')
    def test_populate_stub(self, mock_fetch):
        mock_fetch.return_value = {
            'id': 1,
            'product_line': 'PPP',
            'href': '/1',
            'departures_start_date': '2013-01-01',
            'departures_end_date': '2014-01-01',
            'tour_dossier': {
                'id': 1,
            },
            'departures': [
                {'id': 1},
                {'id': 2},
            ],
        }
        data = {'id': 1, 'product_line': 'PPP'}

        t = Tour(data, stub=True)
        self.assertTrue(t.is_stub)
        self.assertEquals(t.id, 1)

        self.assertTrue(isinstance(t.tour_dossier, TourDossier))
        self.assertTrue(isinstance(t.departures, Query))

        # Force a fetch.
        assert t.departures_start_date

        mock_fetch.assert_called_once()
        self.assertFalse(t.is_stub)
        self.assertEquals(
            t.departures_start_date,
            datetime.datetime.strptime('2013-01-01', DATE_FORMAT).date()
        )
        self.assertEquals(
            t.departures_end_date,
            datetime.datetime.strptime('2014-01-01', DATE_FORMAT).date()
        )

    def test_model_fields(self):
        from gapipy.models.base import BaseModel

        class Bar(BaseModel):
            _as_is_fields = ['id']
            _date_fields = ['date']

        class Foo(Resource):
            _model_fields = [('bar', Bar)]

        data = {
            'bar': {
                'id': 1,
                'date': '2013-01-01',
            }
        }
        f = Foo(data)

        self.assertEquals(f.to_dict(), {
            'bar': {'id': 1, 'date': '2013-01-01'}
        })
        self.assertIsInstance(f.bar, Bar)
        self.assertEquals(f.bar.id, 1)
        self.assertEquals(f.bar.date, datetime.date(2013, 01, 01))

    def test_null_model_fields(self):
        from gapipy.models.base import BaseModel

        class Bar(BaseModel):
            _as_is_fields = ['id']
            _date_fields = ['date']

        class Foo(Resource):
            _model_fields = [('bar', Bar)]

        data = {
            'bar': None
        }
        f = Foo(data)

        self.assertEquals(f.bar, None)

class AccommodationRoomTestCase(TestCase):

    def test_room_class_is_set_properly(self):
        data = {
            'code': 1234,
            'min_nights': 1,
            'max_nights': 2,
            'price_bands': [],
            'class': 'Standard',
        }
        room = AccommodationRoom(data)
        self.assertEqual(room.room_class, 'Standard')

    def test_room_class_is_set_properly_when_cached(self):
        data = {
            'code': 1234,
            'min_nights': 1,
            'max_nights': 2,
            'price_bands': [],
            'room_class': 'Standard',
        }
        room = AccommodationRoom(data)
        self.assertEqual(room.room_class, 'Standard')

class PromotionTestCase(TestCase):
    @patch('gapipy.request.APIRequestor._request')
    def test_product_type_is_set_properly(self, mock_request):
        data = {
            'products': [
                {
                    'type': 'departures',
                    'sub_type': 'Tour',
                    'id': 1,
                }
            ],
        }

        promotion = Promotion(data)
        product = promotion.products[0]
        self.assertEqual(product.type, 'departures')
        self.assertEqual(product.to_dict(), {
            'id': 1,
            'type': 'departures',
            'sub_type': 'Tour',
        })

        # Ensure when the product implicitly fetched it makes a request, and
        # retains the type fields.
        self.assertEqual(mock_request.call_count, 0)
        mock_request.return_value = {
            'finish_address': {
                'country': {
                    'name': 'Australia',
                }
            }
        }
        self.assertEqual(product.finish_address.country.name, 'Australia')
        self.assertEqual(product.type, 'departures')



from __future__ import unicode_literals

from ..base import Resource
from .place import Place


class Airport(Resource):

    _resource_name = 'airports'

    _as_is_fields = [
        'id', 'href', 'name', 'iata_code', 'icao_code', 'active'
    ]
    _date_time_fields_utc = ['date_created', 'date_last_modified']
    _resource_fields = [
        ('place', Place),
        ('city', Place),
    ]



from .base import BaseModel

class Room(BaseModel):
    _as_is_fields = ['code', 'name']

class AdvertisedDeparture(BaseModel):
    _as_is_fields = ["previous_amount", "currency", "amount"]
    _model_fields = [
        ('room', Room),
    ]

    @property
    def _resource_fields(self):
        # Import loop
        from ..resources import Departure, Promotion
        return [
            ('departure', Departure),
            ('promotion', Promotion),
        ]



from __future__ import unicode_literals


from ..base import Product
from ..dossier import ActivityDossier


class Activity(Product):

    _resource_name = 'activities'
    _is_listable = False

    _as_is_fields = [
        'id', 'href', 'availability', 'name', 'description', 'product_line',
        'sku', 'type', 'sub_type', 'requirements',
    ]
    _date_fields = ['start_date', 'finish_date']
    _date_time_fields_utc = ['date_created', 'date_last_modified']

    @property
    def _model_collection_fields(self):
        from ...models import PriceBand
        return [
            ('price_bands', PriceBand),
        ]

    _resource_fields = [('activity_dossier', ActivityDossier)]



from .base import BaseModel


class TravellerHeight(BaseModel):
    _as_is_fields = ['height']

    @property
    def _resource_fields(self):
        from gapipy.resources import Customer
        return [
            ('customer', Customer),
        ]



from .base import BaseModel


class DepartureServiceRoom(BaseModel):
    _as_is_fields = ['name', 'code']

    @property
    def _resource_fields(self):
        from gapipy.resources import Customer
        return [
            ('customers', Customer),
        ]



from .addon import AddOn
from .price_band import PriceBand, SeasonalPriceBand
from .base import BaseModel


class Room(BaseModel):
    _as_is_fields = ['availability', 'code', 'name']

    @property
    def _model_collection_fields(self):
        return [('price_bands', PriceBand)]

    def __repr__(self):
        return '<{0} ({1})>'.format(self.__class__.__name__, self.name)


class AccommodationRoom(Room):

    @property
    def _as_is_fields(self):
        return super(AccommodationRoom, self)._as_is_fields + [
            'room_class',
            'max_nights',
            'min_nights',
        ]

    @property
    def _model_collection_fields(self):
        return [('price_bands', SeasonalPriceBand)]

    def __init__(self, data):
        super(AccommodationRoom, self).__init__(data)

        # `class` is a reserved word in python, so we instead use `room_class`
        # as the attribute for the room class.

        room_class = data.get('class')
        if room_class is None:
            # Sometimes, the data we receive comes from `gapipy` (for example,
            # if local caching is used), so the info we need is already at the
            # `room_class` key.
            room_class = data.get('room_class')

        setattr(self, 'room_class', room_class)


class DepartureRoom(Room):
    @property
    def _as_is_fields(self):
        return super(DepartureRoom, self)._as_is_fields + [
            'flags',
        ]

    @property
    def _model_collection_fields(self):
        return super(DepartureRoom, self)._model_collection_fields + [
            ('addons', AddOn),
        ]



from __future__ import unicode_literals

from ..base import Resource
from .feature_category import FeatureCategory


class Feature(Resource):

    _resource_name = 'features'

    _as_is_fields = [
        'id', 'href', 'name', 'code', 'description',
    ]
    _date_time_fields_utc = ['date_created', 'date_last_modified']
    _resource_fields = [
        ('feature_category', FeatureCategory),
    ]



# flake8: NOQA
from .agency import Agency
from .agent import Agent
from .booking import Booking
from .customer import Customer
from .declined_reason import DeclinedReason
from .document import Invoice, Document
from .nationality import Nationality
from .service import (
    AccommodationService,
    ActivityService,
    DepartureService,
    FeeService,
    FlightService,
    InsuranceService,
    SingleSupplementService,
    TransportService,
)
from .transaction import Payment, Refund



from decimal import Decimal
from itertools import ifilter, ifilterfalse
import datetime

from gapipy import client as client_module
from gapipy.query import Query
from gapipy.utils import (
    get_resource_class_from_class_name,
    get_resource_class_from_resource_name,
)

DATE_FORMAT = '%Y-%m-%d'
DATE_TIME_UTC_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
DATE_TIME_LOCAL_FORMAT = "%Y-%m-%dT%H:%M:%S"


class BaseModel(object):
    """
    A base for any Resource or model type. The difference being a Resource
    points to a specific API resource while a model type might have a reference
    to a resource but is mostly nested type that is complex and has attributes
    adjusted for it.
    """
    _as_is_fields = []
    _date_fields = []
    _date_time_fields_utc = []
    _date_time_fields_local = []
    _model_fields = []
    _model_collection_fields = []
    _price_fields = []
    _resource_fields = []
    _resource_collection_fields = []
    _deprecated_fields = []

    def __init__(self, data, client=None):
        self._client = client or client_module.current_client
        self._raw_data = data
        self._fill_fields(data)

    def _fill_fields(self, data):
        self._raw_data = data
        first = lambda l: [pair[0] for pair in l]

        # Initially we populate base fields, as model/resource fields may rely
        # on these to be present.
        remaining_data = {}
        for field, value in data.iteritems():
            if field in self._as_is_fields:
                self._set_as_is_field(field, value)
            elif field in self._date_fields:
                self._set_date_field(field, value)
            elif field in self._date_time_fields_local:
                self._set_date_time_field_local(field, value)
            elif field in self._date_time_fields_utc:
                self._set_date_time_field_utc(field, value)
            elif field in self._price_fields:
                self._set_price_field(field, value)
            else:
                remaining_data[field] = value

        # Populate resource/model fields.
        for field, value in remaining_data.iteritems():
            if field in first(self._model_fields):
                self._set_model_field(field, value)
            elif field in first(self._model_collection_fields):
                self._set_model_collection_field(field, value)
            elif field in first(self._resource_fields):
                self._set_resource_field(field, value)
            elif field in first(self._resource_collection_fields):
                self._set_resource_collection_field(field, value)

    def _set_as_is_field(self, field, value):
        setattr(self, field, value)

    def _set_date_field(self, field, value):
        if value:
            value = datetime.datetime.strptime(value, DATE_FORMAT).date()
        setattr(self, field, value)

    def _set_date_time_field_local(self, field, value):
        if value:
            value = datetime.datetime.strptime(value, DATE_TIME_LOCAL_FORMAT)
        setattr(self, field, value)

    def _set_date_time_field_utc(self, field, value):
        if value:
            value = datetime.datetime.strptime(value, DATE_TIME_UTC_FORMAT)
        setattr(self, field, value)

    def _set_price_field(self, field, value):
        if value:
            value = Decimal(value)
        setattr(self, field, value)

    def _model_cls(self, field):
        # Find what model class this field is in reference to by plucking it
        # from the set of all fields that allow this type of definition.
        fields = (self._model_fields
                  + self._model_collection_fields
                  + self._resource_fields
                  + self._resource_collection_fields)

        model_cls = [cls for f, cls in fields if f == field][0]

        # FIXME: This will not work for the model_*_fields.
        if isinstance(model_cls, basestring):
            model_cls = get_resource_class_from_class_name(model_cls)
        return model_cls

    def _set_model_field(self, field, value):
        if value is None:
            setattr(self, field, None)
        else:
            setattr(self, field, self._model_cls(field)(value))

    def _set_model_collection_field(self, field, value):
        model_cls = self._model_cls(field)
        items = [model_cls(m) for m in value]
        setattr(self, field, items)

    def _set_resource_field(self, field, value):
        stub = None

        if isinstance(value, list):
            stub = []
            for v in value:
                stub.append(self._model_cls(field)(v, stub=True))
        elif value:
            stub = self._model_cls(field)(value, stub=True)
        setattr(self, field, stub)

    def _set_resource_collection_field(self, field, value):
        is_parent_resource = getattr(self, '_is_parent_resource', None)
        if is_parent_resource:
            parent = (self._resource_name, self.id)
        else:
            parent = None
        raw_data = value
        query = Query(self._client, self._model_cls(field), parent=parent, raw_data=raw_data)
        setattr(self, field, query)

    def _allowed_fields(self):
        first = lambda pair: pair[0]
        return (
            self._as_is_fields
            + self._date_fields
            + self._date_time_fields_utc
            + self._date_time_fields_local
            + map(first, self._model_fields)
            + map(first, self._model_collection_fields)
            + self._price_fields
            + map(first, self._resource_fields)
            + map(first, self._resource_collection_fields)
            + self._deprecated_fields
        )

    def _convert_from_resource_type(self, key, value):
        # Convert instance values into serializable objects.
        if isinstance(value, BaseModel):
            return value.to_dict()
        elif isinstance(value, Query):
            return value._to_dict()
        elif isinstance(value, Decimal):
            return str(value)
        elif isinstance(value, (datetime.date, datetime.datetime)):
            if key in self._date_fields:
                return datetime.datetime.strftime(value, DATE_FORMAT)
            elif key in self._date_time_fields_utc:
                return datetime.datetime.strftime(value, DATE_TIME_UTC_FORMAT)
            elif key in self._date_time_fields_local:
                return datetime.datetime.strftime(value, DATE_TIME_LOCAL_FORMAT)
        else:
            return value

    def to_dict(self):
        properties = {k: v for k, v in self.__dict__.iteritems() if k in self._allowed_fields()}
        data = {}
        for key, value in properties.iteritems():
            if isinstance(value, (list, tuple)):
                data[key] = [self._convert_from_resource_type(key, a) for a in value]
            else:
                data[key] = self._convert_from_resource_type(key, value)
        return data


class RelatedResourceMixin(object):
    _related_resource_lookup = None

    def __init__(self, data):
        super(RelatedResourceMixin, self).__init__(data)

        assert self._related_resource_lookup is not None

        resource_cls = get_resource_class_from_resource_name(
            getattr(self, self._related_resource_lookup))
        resource = resource_cls({'id': self.id, 'href': self.href}, stub=True)
        setattr(self, 'resource', resource)

class DictToModel(object):
    """
    A simple container that is used by `utils.dict_to_model` to help create
    dot-accessible models without being explicit about field definitions.

    This class simply iterates over any complex values and recursively creates
    new objects for those their respective keys, allowing the entire path to be
    dot-accessible.
    """
    def __init__(self, data, class_name=None):
        self._class_name = self._humanize(class_name) if class_name else ''

        # Add any shallow data to this object as primitive types.
        shallow = self._shallow(data)
        self.__dict__.update(shallow)

        # Anything that will contain nested values that need to be dot
        # accessible receive treatment of having its own class representation.
        for k, v in self._deep(data).items():
            if isinstance(v, (list, tuple)):
                value = [DictToModel(i, class_name=k) for i in v]
            else:
                value = DictToModel(v, class_name=k)
            setattr(self, k, value)

    def __str__(self):
        return self._class_name

    def __repr__(self):
        return '<{}>'.format(self.__str__())

    def _humanize(self, class_name):
        return class_name.replace("_", " ").title()

    def _get_data(self, data):
        """ Use as predicate for _shallow, _deep """
        return isinstance(data[1], (dict, list))

    def _shallow(self, data):
        return {k: v for k, v in ifilterfalse(self._get_data, data.items())}

    def _deep(self, data):
        return {k: v for k, v in ifilter(self._get_data, data.items())}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import gapipy

from setuptools import setup, find_packages


readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

requirements = [
    'requests',
]

test_requirements = [
    'mock',
    'nose',
    'nose-parameterized',
]

setup(
    name='gapipy',
    version=gapipy.__version__,
    description='Python client for the G Adventures REST API',
    long_description=readme + '\n\n' + history,
    author='G Adventures',
    author_email='software@gadventures.com',
    url='https://github.com/gadventures/gapipy',
    packages=find_packages(),
    package_dir={'gapipy': 'gapipy'},
    include_package_data=True,
    install_requires=requirements,
    license='MIT',
    zip_safe=False,
    keywords='gapipy',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
    test_suite='nose.collector',
    tests_require=test_requirements,
)



from __future__ import unicode_literals

from ..base import Resource


class BaseTransaction(Resource):
    _is_listable = False

    _as_is_fields = ['id', 'href', 'status', 'transaction_type', 'error']
    _date_time_fields_utc = ['date_created', 'date_last_modified']
    _price_fields = ['amount']
    _resource_fields = [
        ('booking', 'Booking'),
    ]


class Refund(BaseTransaction):
    _resource_name = 'refunds'
    _price_fields = ['amount', 'commission']


class Payment(BaseTransaction):
    _resource_name = 'payments'



from __future__ import unicode_literals

from ..base import Resource
from .details import DossierDetail, DossierDetailsMixin


class TransportDossier(Resource, DossierDetailsMixin):
    _resource_name = 'transport_dossiers'

    _as_is_fields = [
        'id', 'href', 'features', 'capacity', 'private', 'name',
        'dossier_segment',
    ]

    _model_collection_fields = [
        ('details', DossierDetail),
    ]

    _date_time_fields_local = ['date_created', 'date_last_modified']



from __future__ import unicode_literals

from ...models import Address

from ..base import Product
from ..dossier import AccommodationDossier


class Accommodation(Product):

    _resource_name = 'accommodations'
    _is_listable = False

    _as_is_fields = ['id', 'href', 'name', 'product_line', 'sku', 'type',
        'sub_type', 'phone_numbers']
    _date_time_fields_utc = ['date_created', 'date_last_modified']
    _model_fields = [('address', Address)]

    @property
    def _model_collection_fields(self):
        from ...models import AccommodationRoom
        return [
           ('rooms', AccommodationRoom), 
        ]

    @property
    def _resource_fields(self):
        return [
           ('accommodation_dossier', AccommodationDossier),
        ]

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.name)

